import unittest

import grpc

from . import helloworld_pb2_grpc
from .app import app
from .helloworld_pb2 import HelloRequest
from ...homi.test_case import HomiRealServerTestCase


class AltsTestCase(HomiRealServerTestCase):
    app = app
    alts = True

    # def test_hello_say(self):
    #
    #     channel = grpc.secure_channel(f'{self.default_server_config["host"]}:{self.default_server_config["port"]}',
    #                                   self.channel_credentials)
    #
    #     stub = helloworld_pb2_grpc.GreeterStub(channel)
    #     name = 'tom'
    #     response = stub.SayHello(HelloRequest(name=name))
    #
    #     self.assertEqual(response.message, f'Hello {name}!')
    #
    # def test_hello_say_group(self):
    #     channel = grpc.secure_channel(f'{self.default_server_config["host"]}:{self.default_server_config["port"]}',
    #                                   self.channel_credentials)
    #
    #     stub = helloworld_pb2_grpc.GreeterStub(channel)
    #     name = "groupA"
    #     response = list(stub.SayHelloGroup(HelloRequest(name=name)))
    #
    #     self.assertEqual(len(response), 4)
    #     return_names = ['a', 'b', 'c', 'd']
    #     for idx, rep in enumerate(response):
    #         self.assertEqual(rep.message, f"Hello {return_names[idx]}!")
    #
    # def test_hello_everyone(self):
    #     channel = grpc.secure_channel(f'{self.default_server_config["host"]}:{self.default_server_config["port"]}',
    #                                   self.channel_credentials)
    #
    #     stub = helloworld_pb2_grpc.GreeterStub(channel)
    #     names = ["tom", 'sam', 'wony', 'homi']
    #
    #     response = stub.HelloEveryone((HelloRequest(name=name) for name in names))
    #     self.assertEqual(response.message, f'Hello everyone {names}!')

    # def test_say_hello_one_by_one(self):
    #     channel = grpc.secure_channel(f'{self.default_server_config["host"]}:{self.default_server_config["port"]}',
    #                                   self.channel_credentials)
    #
    #     stub = helloworld_pb2_grpc.GreeterStub(channel)
    #     names = ["tom", 'sam', 'wony', 'homi']
    #
    #     response = list(stub.SayHelloOneByOne((HelloRequest(name=name) for name in names)))
    #
    #     self.assertEqual(len(response), len(names))
    #
    #     for rep, name in zip(response, names):
    #         self.assertEqual(rep.message, f'Hello {name}!')

    # def test_fail_try_connect_insecure_channel(self):
    #     channel = grpc.insecure_channel(f'{self.default_server_config["host"]}:{self.default_server_config["port"]}')
    #
    #     stub = helloworld_pb2_grpc.GreeterStub(channel)
    #     name = 'tom'
    #     with self.assertRaises(grpc._channel._InactiveRpcError) as _:
    #         stub.SayHello(HelloRequest(name=name))

    def test_fail_try_wrong_credentials(self):
        cert = b'abcd'
        wrong_credentials = grpc.ssl_channel_credentials(cert)
        channel = grpc.secure_channel(f'{self.default_server_config["host"]}:{self.default_server_config["port"]}',
                                      wrong_credentials)
        stub = helloworld_pb2_grpc.GreeterStub(channel)
        name = 'tom'
        with self.assertRaises(grpc._channel._InactiveRpcError) as _:
            stub.SayHello(HelloRequest(name=name))


if __name__ == '__main__':
    unittest.main()



from .helloworld_pb2 import _GREETER, HelloReply
from ...homi import App

app = App(
    services=[
        _GREETER,
    ]
)
service_name = 'helloworld.Greeter'


# unary-unary method
@app.method(service_name)
def SayHello(name, **kwargs):
    print(f"{name} is request SayHello")
    return HelloReply(message=f"Hello {name}!")


# unary-stream method
@app.method(service_name)
def SayHelloGroup(name, **kwargs):
    print(f"{name} is request SayHelloGroup")
    names = ['a', 'b', 'c', 'd']
    for name in names:
        yield HelloReply(message=f"Hello {name}!")


# stream-unary method
@app.method(service_name)
def HelloEveryone(request_iterator, context):
    names = []
    for reqs in request_iterator:
        print('you can get raw request', reqs.raw_data)
        names.append(reqs['name'])
    return HelloReply(message=f"Hello everyone {names}!")


# stream-stream method
@app.method(service_name)
def SayHelloOneByOne(request_iterator, context):
    for req in request_iterator:
        name = req['name']
        print(f"{name} say to you hello")
        yield HelloReply(message=f"Hello {name}!")



from abc import ABC


class BaseServiceConfig(ABC):
    def __init__(self, name: str, default: dict = None):
        self._default = default
        self.name = name
        self.app = None

    def register_app(self, app):
        self.app = app

    def get_config(self) -> dict:
        return self._default


class MergeConfig(BaseServiceConfig):

    def register_app(self, app):
        super().register_app(app)
        app_config = self.app.config.get(self.name)
        if app_config:
            self._default.update(app_config)



from .app import App, BaseApp, BaseService, Service
from .server import Server
from .aio.app import AsyncApp, BaseAsyncApp, BaseAsyncService, AsyncService
from .aio.server import AsyncServer

__version__ = "0.2.0"



import logging
from abc import ABC, abstractmethod
from typing import Callable, Dict, Generic, List, TypeVar, Union

import grpc
from google.protobuf.descriptor import ServiceDescriptor

from ..app import BaseService
from ..config import MergeConfig
from ..exception import MethodNotFound, RegisterError, ServiceNotFound
from ..proto_meta import ServiceMetaData, make_grpc_method_handler, service_metadata_from_descriptor, warp_async_handler


async def AsyncNotImplementedMethod(request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


class BaseAsyncApp:
    def __init__(self, config: dict = None, **kwargs):
        self._config: dict = config or {}
        pass

    @property
    def config(self):
        return self._config


ConfigType = TypeVar('ConfigType')


class BaseAsyncService(Generic[ConfigType], ABC):
    def __init__(
            self,
            service_descriptor: ServiceDescriptor,
            config_class=MergeConfig,
            default_config: ConfigType = None,
            config_name=None,
            **kwargs
    ):
        self._register_methods = {}
        self._app: BaseAsyncApp = None
        self._is_registered = False

        self._descriptor = service_descriptor
        self._meta: ServiceMetaData = service_metadata_from_descriptor(service_descriptor)
        self._config: config_class

        self._prepare_config(config_class, config_name, default_config)

        # lifecycle method
        self._after_registered_handler = None
        self._before_server_start_handler = None

    @property
    def descriptor(self):
        return self._descriptor

    @property
    def meta(self):
        return self._meta

    @property
    def full_name(self):
        return self.meta.full_name

    @property
    def name(self):
        return self.meta.name

    @property
    def method_names(self):
        return self.meta.methods.keys()

    def _prepare_config(self, klass, name, default):
        config_name = name or self.full_name
        default_config = default or {}
        self._config = klass(config_name, default_config)

    @property
    def config(self) -> ConfigType:
        if not self.is_registered:
            logging.warning('your get config data before extend registered')
        return self._config.get_config()

    @property
    def app(self):
        if not self.is_registered:
            raise ValueError('you can not get app object before extend registered')
        return self._app

    @property
    def is_registered(self) -> bool:
        if not self._is_registered and self._app:
            self._is_registered = True
        return self._is_registered

    def register_app(self, app):
        self._app = app
        self._config.register_app(app)
        if self._after_registered_handler:
            self._after_registered_handler(self)

            # todo: if register app can request async remove annotation
            # if inspect.isawaitable(self._after_registered_handler):
            #     self._after_registered_handler(self)
            # else:
            #     self._after_registered_handler(self)

    def before_server_start_handler(self):
        if self._before_server_start_handler:
            self._before_server_start_handler(self)

            # todo: if register app can request async remove annotation
            # if inspect.isawaitable(self._before_server_start_handler):
            #     await self._before_server_start_handler(self)
            # else:
            #     self._before_server_start_handler(self)

    def after_registered(self, func):
        self._after_registered_handler = func
        return func

    def before_server_start(self, func):
        self._before_server_start_handler = func
        return func

    @abstractmethod
    def add_to_server(self, server):
        raise NotImplementedError('YOU MUST OVERWRITE THIS METHOD!!')
        pass


class AsyncService(Generic[ConfigType], BaseAsyncService[ConfigType]):

    def __init__(self, service_descriptor: ServiceDescriptor, config_class=MergeConfig, default_config=None,
                 config_name=None, **kwargs):
        super().__init__(service_descriptor, config_class, default_config, config_name, **kwargs)

        self._method_handler: Dict[str, Callable] = {}

    # method func register decorator
    def method(self, method_name=None, **kwargs):
        def wrapped(func: Callable):
            name = method_name or func.__name__
            if name not in self.meta.methods:
                raise MethodNotFound(name, self.meta.full_name, available_methods=self.meta.methods.keys())
            self._method_handler[name] = func
            return func

        return wrapped

    def register_method(self, method_name, func: Callable, **kwargs):
        self.method(method_name, **kwargs)(func)

    def make_servicer_class(self):
        methods = {}
        for name, method_meta in self.meta.methods.items():
            if name in self._method_handler:
                func = warp_async_handler(method_meta, self._method_handler[name])
            else:
                func = AsyncNotImplementedMethod
            methods[name] = func
        servicer = type(self.full_name, (), methods)
        return servicer

    def _make_generic_handler(self) -> Dict[str, Callable]:
        generic_handler = {}
        for name, method_meta in self.meta.methods.items():
            if name in self._method_handler:
                func = warp_async_handler(method_meta, self._method_handler[name])
            else:
                func = AsyncNotImplementedMethod
            generic_handler[name] = make_grpc_method_handler(method_meta, func, )
        return grpc.method_handlers_generic_handler(self.full_name, generic_handler)

    def add_to_server(self, server):
        server.add_generic_rpc_handlers((self._make_generic_handler(),))


class AsyncApp(BaseAsyncApp):
    def __init__(self, services: List[Union[AsyncService, ServiceDescriptor]] = None, **kwargs):
        super().__init__(**kwargs)

        self._services: Dict[str, AsyncService] = {}

        _services = services or []
        for svc in _services:
            if isinstance(svc, ServiceDescriptor):
                self.add_service_from_descriptor(svc)
            else:
                self.add_service(svc)

    @property
    def service_names(self):
        return self._services.keys()

    @property
    def services(self):
        return self._services.values()

    def get_service_by_full_name(self, full_name: str) -> AsyncService:
        try:
            return self._services[full_name]
        except KeyError:
            raise ServiceNotFound(full_name, available_services=self.service_names)

    def add_service(self, service: AsyncService):
        if not isinstance(service, BaseAsyncService) and not isinstance(service, BaseService):
            raise RegisterError('you must register Service class object')
        self._services[service.full_name] = service
        service.register_app(self)

    def add_service_from_descriptor(self, descriptor: ServiceDescriptor):
        self.add_service(AsyncService(descriptor))

    def method(self, service_full_name: str, method_name: str = None, **kwargs):
        svc = self.get_service_by_full_name(service_full_name)

        def wrapped(func: Callable):
            name = method_name or func.__name__
            svc.register_method(name, func, **kwargs)
            return func

        return wrapped

    def register_method(self, service_full_name: str, method_name: str, func: Callable, **kwargs):
        self.method(service_full_name, method_name, **kwargs)(func)

    def bind_to_server(self, server):
        for svc in self.services:
            svc.before_server_start_handler()

        for svc in self.services:
            svc.add_to_server(server)



import os
from concurrent import futures

import grpc

from .app import App
from .exception import ServerSSLConfigError


class Server:
    def __init__(self,
                 app: App,
                 host: str = None,
                 port: str = '50051',
                 worker: int = 10,
                 debug: bool = False,
                 alts: bool = False,
                 private_key: bytes = None,
                 certificate: bytes = None,
                 ):
        self.host = host
        self.port = port
        self.worker = worker
        self.app = app
        self.debug = debug
        self.load_config_from_env()
        self.alts = alts
        self.private_key = private_key
        self.certificate = certificate
        self.server = None
        self._server_credentials = None
        self.thread_pool = futures.ThreadPoolExecutor(max_workers=self.worker)

    def load_config_from_env(self):
        pass

    @property
    def endpoint(self):
        host = self.host if self.host else "[::]"
        return f'{host}:{self.port}'

    @property
    def server_credentials(self):
        if not self._server_credentials:
            self._server_credentials = grpc.ssl_server_credentials(
                ((self.private_key, self.certificate,),))
        return self._server_credentials

    def _add_port(self):
        if self.alts:
            self.server.add_secure_port(self.endpoint, grpc.alts_server_credentials())
        elif self.private_key and self.certificate:
            self.server.add_secure_port(self.endpoint, self.server_credentials)
        elif self.private_key or self.certificate:
            # error
            raise ServerSSLConfigError('If you want enable ssl feature, You Must set tls_key, certificate config both ')
        else:
            self.server.add_insecure_port(self.endpoint)

    def run(self, wait=True):
        if self.debug:
            os.environ['GRPC_VERBOSITY'] = 'debug'
        self.server = grpc.server(self.thread_pool)
        self.app.bind_to_server(self.server)
        self._add_port()
        self.server.start()
        print('run server')
        print(f'# port : {self.port}')
        if wait:
            self.wait_for_termination()

    def stop(self, grace=None):
        if self.server:
            self.server.stop(grace)

    def wait_for_termination(self):
        if self.server:
            self.server.wait_for_termination()



from enum import Enum
from functools import partial
from inspect import signature
from typing import Any, Dict, NamedTuple, Tuple, TypeVar

import grpc
from google.protobuf import json_format, symbol_database
from google.protobuf.descriptor import MethodDescriptor, ServiceDescriptor
from google.protobuf.descriptor_pb2 import MethodDescriptorProto, ServiceDescriptorProto


class MethodType(Enum):
    UNARY_UNARY = 'unary_unary'
    STREAM_UNARY = 'stream_unary'
    UNARY_STREAM = 'unary_stream'
    STREAM_STREAM = 'stream_stream'

    @property
    def is_unary_request(self):
        return 'unary_' in self.value

    @property
    def is_unary_response(self):
        return '_unary' in self.value


IS_REQUEST_STREAM = TypeVar("IS_REQUEST_STREAM")
IS_RESPONSE_STREAM = TypeVar("IS_RESPONSE_STREAM")

MethodTypeMatch: Dict[Tuple[IS_REQUEST_STREAM, IS_RESPONSE_STREAM], MethodType] = {
    (False, False): MethodType.UNARY_UNARY,
    (True, False): MethodType.STREAM_UNARY,
    (False, True): MethodType.UNARY_STREAM,
    (True, True): MethodType.STREAM_STREAM,
}


class MethodMetaData(NamedTuple):
    name: str
    input_type: Any
    output_type: Any
    method_type: MethodType


class ServiceMetaData(NamedTuple):
    full_name: str
    name: str
    methods: Dict[str, MethodMetaData]


def get_method_metadata(descriptor: MethodDescriptor, proto: MethodDescriptorProto) -> MethodMetaData:
    symbol_db = symbol_database.Default()
    return MethodMetaData(
        name=descriptor.name,
        input_type=symbol_db.GetPrototype(descriptor.input_type),
        output_type=symbol_db.GetPrototype(descriptor.output_type),
        method_type=MethodTypeMatch[(proto.client_streaming, proto.server_streaming)],
    )


def service_metadata_from_descriptor(service_descriptor: ServiceDescriptor) -> ServiceMetaData:
    svc_desc_proto = ServiceDescriptorProto()
    service_descriptor.CopyToProto(svc_desc_proto)
    methods = {
        proto.name: get_method_metadata(service_descriptor.methods_by_name[proto.name], proto)
        for proto in svc_desc_proto.method
    }

    return ServiceMetaData(
        full_name=service_descriptor.full_name,
        name=service_descriptor.name,
        methods=methods
    )


def make_grpc_method_handler(method_meta: MethodMetaData, func):
    handler = getattr(grpc, f"{method_meta.method_type.value}_rpc_method_handler")
    return handler(
        func,
        request_deserializer=method_meta.input_type.FromString,
        response_serializer=method_meta.output_type.SerializeToString,
    )


class StreamMessage(Dict):
    _raw_data = None


def parse_request(parameters, request) -> Dict:
    request_dict = json_format.MessageToDict(request, preserving_proto_field_name=True)
    args = {}
    for p in parameters:
        args[p] = request_dict.get(p)
    args['request'] = request
    return args


def parse_stream_request(request_iterator) -> Dict:
    for req in request_iterator:
        msg = StreamMessage(**json_format.MessageToDict(req, preserving_proto_field_name=True))
        msg.raw_data = req
        yield msg


async def parse_async_stream_request(request_iterator):
    async for req in request_iterator:
        msg = StreamMessage(**json_format.MessageToDict(req, preserving_proto_field_name=True))
        msg.raw_data = req
        yield msg


def parse_to_dict(input_type, item):
    return json_format.ParseDict(item, input_type()) if isinstance(item, dict) else item


def parse_stream_return(input_type, items):
    for item in items:
        yield parse_to_dict(input_type, item)


async def parse_async_stream_return(input_type, items):
    async for item in items:
        yield parse_to_dict(input_type, item)


def warp_handler(method_meta: MethodMetaData, func):
    sig = signature(func)
    parameters = tuple(k for k, v in sig._parameters.items() if v.kind.value == 1)

    if method_meta.method_type.is_unary_response:
        return_func = partial(parse_to_dict, method_meta.output_type)
    else:
        return_func = partial(parse_stream_return, method_meta.output_type)

    if method_meta.method_type.is_unary_request:
        request_parser = partial(parse_request, parameters)

        def wrapper(request, context):
            result = func(**request_parser(request), context=context)
            return return_func(result)
    else:
        def wrapper(request, context):
            result = func(parse_stream_request(request), context=context)
            return return_func(result)

    return wrapper


def warp_async_handler(method_meta: MethodMetaData, func):
    sig = signature(func)
    parameters = tuple(k for k, v in sig._parameters.items() if v.kind.value == 1)

    is_unary_response = method_meta.method_type.is_unary_response
    is_unary_request = method_meta.method_type.is_unary_request

    if is_unary_response:
        return_func = partial(parse_to_dict, method_meta.output_type)
    else:
        return_func = partial(parse_async_stream_return, method_meta.output_type)

    if is_unary_request:
        request_parser = partial(parse_request, parameters)
        if is_unary_response:
            async def wrapper(request, context):
                result = func(**request_parser(request), context=context)
                return return_func(await result)
        else:
            async def wrapper(request, context):
                result = func(**request_parser(request), context=context)
                async for msg in return_func(result):
                    yield msg

    else:
        if is_unary_response:
            async def wrapper(request, context):
                result = func(parse_async_stream_request(request), context=context)
                return return_func(await result)
        else:
            async def wrapper(request, context):
                result = func(parse_async_stream_request(request), context=context)
                async for msg in return_func(result):
                    yield msg

    return wrapper


def warp_handler_for_method(method_meta: MethodMetaData, func):
    handler = warp_handler(method_meta, func)

    def decorator(request, context):
        return handler(request, context)

    return decorator



import asyncio
import sys
from os.path import dirname

import click

from homi import AsyncServer
from . import AsyncApp
from .exception import ServerSSLConfigError


@click.group()
def cli():
    pass


@click.command("run", short_help="Run a development server.")
@click.argument('file', type=click.Path(exists=True, resolve_path=True), default="app.py")
@click.option("--host", "-h", help="The interface to bind to.")
@click.option("--port", "-p", default='50051', help="The port to bind to.")
@click.option('--worker', '-w', default=10, type=int)
@click.option('--use_uvloop', default=True, type=bool, help="If you don't want uvloop `--uvloop false`")
@click.option('--alts', type=bool, default=False, help='[Experimental] enable alts')
@click.option('--private_key', '-k', type=click.Path(exists=True, resolve_path=True), help='tls private key')
@click.option('--certificate', '-c', type=click.Path(exists=True, resolve_path=True), help='tls root certificate')
# @click.option(
#     "--reload/--no-reload",
#     default=None,
#     help="Enable or disable the reloader. By default the reloader "
#          "is active if debug is enabled.",
# )
@click.option(
    "--debug",
    default=False,
    type=bool,
    help="Server Debug Mode",
)
def run_command(file, port, worker, debug, alts, host=None, private_key=None, certificate=None, use_uvloop=True):
    sys.path.append(dirname(file))
    import importlib.util

    spec = importlib.util.spec_from_file_location("app", file)
    app_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(app_module)

    from .server import Server
    if private_key and certificate:
        with open(private_key, 'rb') as f:
            private_key = f.read()
        with open(certificate, 'rb') as f:
            certificate = f.read()
    elif private_key or certificate:
        raise ServerSSLConfigError('if you want use tls mode, you must set both private_key & certificate value')
    if isinstance(app_module.app, AsyncApp):
        if use_uvloop:
            import uvloop
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        else:
            print('use asyncio loop (uvloop off)')
        server = AsyncServer(
            app_module.app,
            host, port,
            debug=debug,
            alts=alts,
            private_key=private_key,
            certificate=certificate
        )
        asyncio.run(server.run())
    else:
        Server(
            app_module.app,
            host, port, worker,
            debug=debug,
            alts=alts,
            private_key=private_key,
            certificate=certificate
        ).run()


cli.add_command(run_command)

if __name__ == '__main__':
    cli()



from typing import TypedDict, Optional, List

from grpc_health.v1 import health, health_pb2_grpc
from grpc_health.v1.health_pb2 import _HEALTH, HealthCheckResponse
from grpc_reflection.v1alpha import reflection
from grpc_reflection.v1alpha.reflection_pb2 import _SERVERREFLECTION

from ..app import BaseService


class ReflectionConfig(TypedDict):
    enable: Optional[bool]
    ignore_service: Optional[List[str]]


class ReflectionService(BaseService[ReflectionConfig]):

    def __init__(self, **kwargs):
        service_descriptor = _SERVERREFLECTION
        default_config: ReflectionConfig = ReflectionConfig(
            enable=True,
            ignore_service=[]
        )

        super().__init__(
            service_descriptor=service_descriptor,
            default_config=default_config,
            config_name='reflection',
            **kwargs
        )

    @property
    def reflection_services(self):
        return [name for name in self.app.service_names if name not in self.config['ignore_service']]

    def add_to_server(self, server):
        if self.config['enable']:
            reflection.enable_server_reflection(self.reflection_services, server)


reflection_service = ReflectionService()


class HealthConfig(TypedDict):
    enable: Optional[bool]
    not_serving: Optional[List[str]]


class HealthService(BaseService[HealthConfig]):

    def __init__(self, **kwargs):
        service_descriptor = _HEALTH
        default_config = HealthConfig(
            enable=True,
            not_serving=[]
        )

        super().__init__(
            service_descriptor=service_descriptor,
            default_config=default_config,
            config_name='health',
            **kwargs
        )

    def add_to_server(self, server):
        if self.config['enable']:
            health_servicer = health.HealthServicer(experimental_non_blocking=True)
            health_servicer.set('', HealthCheckResponse.SERVING)
            for svc in self.app.service_names:
                if svc in self.config['not_serving']:
                    status = HealthCheckResponse.NOT_SERVING
                else:
                    status = HealthCheckResponse.SERVING
                health_servicer.set(svc, status)
            health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)


health_service = HealthService()



import asyncio
import datetime
import logging
import threading
import unittest
from asyncio import Queue as AsyncQueue
from time import sleep
from typing import Any, Dict

import grpc
import grpc_testing
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from . import App, AsyncApp, AsyncServer, Server, Service


class HomiTestCase(unittest.TestCase):
    app: App = None
    _test_server = None

    def get_test_server(self):
        if not self._test_server:
            servicers = {}
            for svc in self.app.services:
                if isinstance(svc, Service):
                    servicers[svc.descriptor] = svc.make_servicer_class()

            self._test_server = grpc_testing.server_from_dictionary(
                servicers, grpc_testing.strict_real_time()
            )
        return self._test_server

    @staticmethod
    def get_all_response(method):
        finish = False
        result = []
        while not finish:
            try:
                result.append(method.take_response())
            except Exception:
                finish = True
        return result

    @staticmethod
    def send_request_all(method, requests):
        [method.send_request(req) for req in requests]
        method.requests_closed()


class HomiRealServerTestCase(unittest.TestCase):
    app = None
    default_server_config = {
        "host": "localhost",
        "port": '5999'
    }
    alts = False
    tls = False
    _tls_key = None
    _certificate = None
    test_server_config: Dict[str, Any] = {}
    test_server = None

    @property
    def tls_key(self):
        if not self._tls_key:
            self._tls_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        return self._tls_key

    @property
    def certificate(self):
        if not self._certificate:
            subject = issuer = x509.Name([
                x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
                x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
                x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
                x509.NameAttribute(NameOID.ORGANIZATION_NAME, "My Company"),
                x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
            ])
            self._certificate = x509.CertificateBuilder() \
                .subject_name(subject) \
                .issuer_name(issuer) \
                .public_key(self.tls_key.public_key()) \
                .serial_number(x509.random_serial_number()) \
                .not_valid_before(datetime.datetime.utcnow()) \
                .not_valid_after(
                # Our certificate will be valid for 10 days
                datetime.datetime.utcnow() + datetime.timedelta(days=10)
            ) \
                .add_extension(x509.SubjectAlternativeName([x509.DNSName("localhost")]),
                               critical=False, ).sign(self.tls_key, hashes.SHA256())
        return self._certificate

    @property
    def channel_credentials(self):
        if self.alts:
            return grpc.alts_channel_credentials()
        elif self.tls:
            cert = self.certificate.public_bytes(serialization.Encoding.PEM)
            return grpc.ssl_channel_credentials(cert)
        return None

    @property
    def tls_config(self):
        if not self.tls:
            return {}
        else:
            return {
                "private_key": self.tls_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                ),
                "certificate": self.certificate.public_bytes(serialization.Encoding.PEM),
            }

    def get_server_config(self, merge_config: dict = None):
        config = merge_config or {}

        return {
            **self.default_server_config,
            **self.test_server_config,
            **self.tls_config,
            **config,
        }

    def server_restart(self, merge_config: dict = None):
        self.run_real_server(merge_config)

    def _run_async_server(self, config):
        self.test_server = AsyncServer(self.app, **self.get_server_config(config))
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.que = AsyncQueue()

        def run_server(loop, server, q):
            async def operator(server, q):
                await server.run(wait=False)
                await q.get()
                await server.stop()
                logging.debug('test server stopped')

            loop.run_until_complete(operator(server, q))

        self.thread = threading.Thread(target=run_server, args=(self.loop, self.test_server, self.que))
        self.thread.start()
        # todo: find how to wait until on server
        sleep(0.02)

    def run_real_server(self, merge_config: dict = None):
        config = merge_config or {}
        if self.test_server:
            try:
                self.test_server.stop()
            except Exception:
                pass
        if isinstance(self.app, AsyncApp):
            self._run_async_server(config)
        else:
            self.test_server = Server(self.app, **self.get_server_config(config))
            self.test_server.run(wait=False)

    def setUp(self):
        super().setUp()
        self.run_real_server()

    def server_stop(self):
        if isinstance(self.app, AsyncApp):
            self.loop.call_soon_threadsafe(self.que.put_nowait, 'stop')
            self.thread.join()
            del self.thread
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())
            self.loop.close()
        else:
            self.test_server.stop()

    def tearDown(self):
        super().tearDown()
        self.server_stop()
